use vecmath::vec3_add;

use crate::{primitives::Box3, stl::parser::StlSolid};

/// A node representing a node single in the bounding volume hierarchy
#[derive(Default)]
struct BvhNode {
    aabb: Box3,
    left_child: usize,
    right_child: usize,
    prim_idx: usize,
    prim_count: usize,
}

/// Struct representing the total bounding volume hierarchy
pub struct Bvh {
    nodes: Vec<BvhNode>,
}

/// Return type for get_subdivision_position_and_axis
struct SplitPlane {
    axis: usize,
    pos: f64,
}

impl Bvh {
    /// Create a new Bounding volume hierarchy from an STL solid
    pub fn new(solid: &StlSolid) -> Self {
        // First we need to get all of the triangle centroids
        let centroids: Vec<[f32; 3]> = solid
            .triangles
            .iter()
            .map(|tri| {
                tri.iter()
                    .fold([0.0; 3], |acc, idx| vec3_add(solid.vertices[*idx], acc))
            })
            .collect();

        // Initialize the root node
        let mut nodes = vec![BvhNode {
            prim_idx: 0,
            prim_count: solid.triangles.len(),
            ..Default::default()
        }];
        nodes[0]
            .aabb
            .shrink_wrap_primitives(&solid.vertices, &solid.triangles);

        let mut bvh = Bvh { nodes };

        bvh.subdivide(0);
        bvh
    }

    /// Subdivide a node, subdivision doesn't occur if the node has only 2 primitives
    pub fn subdivide(&mut self, node_idx: usize) {
        let split_plane = self.get_split_plane(node_idx);
    }

    /// Calculate the optimal split plane axis and position.
    /// This function should use the surface area heuristic to select the optimal split
    fn get_split_plane(&self, node_idx: usize) -> SplitPlane {
        // Initial approach just splits halfway along the longest axis
        // TODO: Improve this to use the surface area heuristic
        let axis = (0..3).fold(0, |acc, a| {
            if self.nodes[node_idx].aabb.extents[acc] < self.nodes[node_idx].aabb.extents[a] {
                a
            } else {
                acc
            }
        });

        let pos = self.nodes[node_idx].aabb.extents[axis] / 2.0;
        SplitPlane { axis, pos }
    }
}
