use std::collections::BTreeSet;

use itertools::Itertools;
use vecmath::{vec3_add, vec3_scale};

#[derive(Default)]
pub struct Box3 {
    pub center: [f64; 3],
    pub extents: [f64; 3],
}

impl Box3 {
    /// Create a new box from the bounding values
    pub fn new_from_bounds(min_bound: [f32; 3], max_bound: [f32; 3]) -> Self {
        let mut center = [0.0; 3];
        let mut extents = [0.0; 3];
        for (i, (min_val, max_val)) in min_bound.iter().zip(max_bound).enumerate() {
            center[i] = ((min_val + max_val) / 2.0) as f64;
            extents[i] = (max_val - min_val) as f64;
        }
        Box3 { center, extents }
    }

    /// This function returns the 8 vertices of an axis-aligned bounding box.
    pub fn get_vertices(&self) -> Vec<[f64; 3]> {
        let half_extent = vec3_scale(self.extents, 0.5);

        // Define all eight corners relative to the center.
        [-half_extent[0], half_extent[0]]
            .into_iter()
            .cartesian_product([-half_extent[1], half_extent[1]])
            .cartesian_product([-half_extent[2], half_extent[2]])
            .map(|((x, y), z)| vec3_add(self.center, [x, y, z]))
            .collect()
    }

    /// Expand a box to fit all primitive inside of it.
    /// NOTE: vertices does not need to include only vertices being considered. Instead only
    /// vertices included as indices in the primitives list are considered
    pub fn shrink_wrap_primitives<T>(&mut self, vertices: &[[f32; 3]], primitives: &[T])
    where
        for<'a> &'a T: IntoIterator<Item = &'a usize>,
    {
        // If the current primitive list is empty, zero bounding box and exit
        if primitives.is_empty() {
            self.center = [0.0; 3];
            self.extents = [0.0; 3]
        }
        // Get all of the unique vertex indices so we don't have to check them multiple times
        let unique_vertex_idxs: BTreeSet<&usize> = BTreeSet::from_iter(primitives.iter().flatten());

        // Now expand the box
        let mut min_bd = [f32::INFINITY; 3];
        let mut max_bd = [f32::NEG_INFINITY; 3];

        for idx in unique_vertex_idxs.into_iter() {
            for ((vertex_pos, min_pos), max_pos) in vertices[*idx]
                .iter()
                .zip(min_bd.iter_mut())
                .zip(max_bd.iter_mut())
            {
                *min_pos = vertex_pos.min(*min_pos);
                *max_pos = vertex_pos.max(*max_pos);
            }
        }
    }
}

pub struct Triangle {
    pub v1: [f64; 3],
    pub v2: [f64; 3],
    pub v3: [f64; 3],
}
